unit upronostico;

interface

uses
  Classes, SysUtils, xMatDefs, ufechas, ucosa, Math;

const
  C_DT_1MINUTO = 1.0 / 24.0 / 60.0;

type
  // definición del Cono de Pronósticos para una serie temporal.
  // La "guia" es la especificación del Cono de Pronósticos en el Espacio de la Señal.
  // en base a la guía se calculan el vector "sesgos" y "factor" que contiene
  // que se aplican en el Espacio Gaussiano para guiar la salida del Filtro Lineal.
  // el ruido que se aplica sobre el canal se calcula como:
  // rs = sesgo[j] + factor[j] * Brbg[i]
  // donde "Brbg" es el Ruido Blanco Gaussiano transformado por la matriz B
  // del filtro lineal que corersponde al canal en cuestión
  // El control del cono de pronósticos se realiza mediante una "guia" del cono
  // e indicaciones de cómo abrir el cono de dispersión.
  // El vector "guia" indica valores para la guía. De ese vector, se utilizan los
  // primeros valores para inicializar el estado del filtro lineal.
  // El parámetro NPSA = Número de Pasos Sin Apertura del cono, determina la
  // cantidad de valores de la guía, posteriores a los usados para fijar el estado
  // inicial que serán utilizados como valores determinísticos.
  // Ese valor puede ir entre 0 hasta el total de valores que queden en la guía.
  TPronostico = class(TCosa)

    NRetardos: integer; // Número de valores de la guía que serán usados para
    // inicializar el filtro lineal

    guia: TDAOfNReal; // Guía en el espacio de la señal real
    guia_eg: TDAOfNReal; // Guía en el espacio gasussiano de la señal.

    guia_pe: NReal;      // probabilidad de excedencia de la guía.

    sesgo: TDAOfNReal; // sesgo del ruido (aditivo)
    factor: TDAofNReal; // multiplicadores del ruido (multimplicativo)

    fechaIniSesgo_: TFecha; // fecha a partir de la cual aplica el sesgo

    // parámetros de calibración control del cono de pronóstico
    // a partir de los determinismos.
    NPCC: integer; // Número de Pasos de Control del Cono
    NPLC: integer; // Número de Pasos de Liberación del Cono
    NPSA: integer; // Número de pasos Sin Apertura del Cono (determinista).
    NPAC: integer; // Número de Pasos de Apertura del Cono

    //Cantidad de valores deterministicos usados
    cantValoresDeterministicosUsados: integer;

    constructor Create(capa: integer; xNRetardos, xNPCC, xNPLC, xNPSA,
      xNPAC: integer);
    constructor Create_Default(xNRetardos: integer);
    procedure WriteToText(f: TArchiTexto); override;
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure Free; override;

    function determinismoSoloEstadoInicial: boolean;

    // retorna max( NPCC+NPLC , NPSA+NPAC)
    // es útil para determinar a partir de qué paso ya el cono ya no tiene
    // influencia.
    function NPasosCono: integer;


    // debe ser llamado en el preparar memoria de la fuente
    procedure calcular_rangos(DuracionPasoSorteoEnHoras: NReal);

    // determina si para la fechaDeInicioDelPaso pasada como parámetro es válido el cono
    // depronósticos o no. Si el resultado es TRUE indica que la fecha está en el rango del cono
    // y entonces en kSesgo y kFactor se devuelve los índices para leer los valores correspondientes
    // de los vectores "sesgo" y "factor".
    function fechaEnRango(fechaDeInicioDelPaso: TFecha;
      var kSesgo, kFactor: integer): boolean;

  private
    rangoFechaSesgo: NReal; // rango de aplicación del sesgo.
    rangoFechaFactor: NReal; // rango de aplicación del ruido
  end;


  TPronosticos = class(TListaDeCosas)
    haySesgos: boolean;

    procedure reset_DeterminismosUsados;
    procedure prepararse(DuracionPasoSorteoEnHoras: NReal);
    // retorna el máximo largo en pasos de sorteo de la fuente de los conos
    // de pronósticos represnetado.
    function max_NPasosCono: integer;

    function getItem(i: integer): TPronostico;
    procedure setItem(i: integer; cosa: TPronostico);
    property items[i: integer]: TPronostico read getItem write setItem; default;
  end;

procedure AlInicio;
procedure AlFinal;

implementation


function TPronosticos.getItem(i: integer): TPronostico;
begin
  Result := TPronostico(inherited items[i]);
end;

procedure TPronosticos.setItem(i: integer; cosa: TPronostico);
begin
  inherited items[i] := cosa;
end;


procedure TPronosticos.reset_DeterminismosUsados;
var
  k: integer;
begin
  for k := 0 to Count - 1 do
  begin
    TPronostico(items[k]).cantValoresDeterministicosUsados := 0;
  end;
end;

procedure TPronosticos.prepararse(DuracionPasoSorteoEnHoras: NReal);
var
  k: integer;
begin
  haySesgos := False;
  for k := 0 to Count - 1 do
  begin
    haySesgos := haySesgos or ((items[k].sesgo <> nil) and
      (length(items[k].sesgo) > 0));
    items[k].calcular_rangos(DuracionPasoSorteoEnHoras);
  end;
end;

function TPronosticos.max_NPasosCono: integer;
var
  k: integer;
  m, res: integer;
begin
  res := TPronostico(items[0]).NPasosCono;
  for k := 1 to Count - 1 do
  begin
    m := TPronostico(items[k]).NPasosCono;
    if (m > res) then
      res := m;
  end;
  Result := res;
end;

constructor TPronostico.Create(capa: integer; xNRetardos, xNPCC, xNPLC, xNPSA, xNPAC: integer);
begin
  inherited Create( capa );
  NRetardos := xNRetardos;
  NPCC := xNPCC;
  NPLC := xNPLC;
  NPSA := xNPSA;
  NPAC := xNPAC;

  setlength(guia, NRetardos + NPCC);
  setlength(guia_eg, NRetardos + NPCC);
  setlength(sesgo, NPCC + NPLC);
  setlength(factor, NPSA + NPAC);

  fechaIniSesgo_ := TFecha.Create_Dt(0);
  rangoFechaSesgo := 0;
  guia_pe := 0.5;

  //  cantValoresDeterministicosUsados := 0;
end;

constructor TPronostico.Create_Default(xNRetardos: integer);
begin
  Create(0, xNRetardos, 0, 0, 0, 0);
end;

function TPronostico.NPasosCono: integer;
begin
  Result := max(NPCC + NPLC, NPSA + NPAC);
end;


procedure TPronostico.calcular_rangos(DuracionPasoSorteoEnHoras: NReal);
begin
  self.rangoFechaSesgo := (NPCC + NPLC) * DuracionPasoSorteoEnHoras / 24.0;
  self.rangoFechaFactor := (NPSA + NPAC) * DuracionPasoSorteoEnHoras / 24.0;
end;

function TPronostico.fechaEnRango(fechaDeInicioDelPaso: TFecha;
  var kSesgo, kFactor: integer): boolean;
var
  dxdt: NReal;
begin
  dxdt := fechaDeInicioDelPaso.dt - fechaIniSesgo_.dt + C_DT_1MINUTO;
  if (dxdt > 0) then
  begin
    if (dxdt < rangoFechaSesgo) then
      kSesgo := trunc(dxdt / rangoFechaSesgo * length(sesgo))
    else
      kSesgo := -1; // indicamos fuera de rango.
    if (dxdt < rangoFechaFactor) then
      kFactor := trunc(dxdt / rangoFechaFactor * length(factor))
    else
      kFactor := -1; // indicamos fuera de rango.
  end
  else
  begin
    kSesgo := -1;
    kFactor := -1;
  end;
  Result := (kSesgo >= 0) or (kFactor >= 0);
end;



procedure TPronostico.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('nretardos', nretardos);
  f.wr('guia', guia);
  f.wr('guia_eg', guia_eg);
  f.wr('guia_pe', guia_pe);
  f.wr('NPCC', NPCC);
  f.wr('NPLC', NPLC);
  f.wr('NPSA', NPSA);
  f.wr('NPAC', NPAC);
  f.wr('sesgo', sesgo);
  f.wr('factor', factor);
  f.wr('fechaIniSesgo', fechaIniSesgo_);
  f.wr('rangoFechaSesgo', rangoFechaSesgo);
end;

constructor TPronostico.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('nretardos', nretardos);
  f.rd('guia', guia);
  if f.version >= 82 then
    f.rd('guia_eg', guia_eg);
  f.rd('guia_pe', guia_pe);
  f.rd('NPCC', NPCC);
  f.rd('NPLC', NPLC);
  f.rd('NPSA', NPSA);
  f.rd('NPAC', NPAC);
  f.rd('sesgo', sesgo);
  f.rd('factor', factor);
  f.rd('fechaIniSesgo', fechaIniSesgo_);
  f.rd('rangoFechaSesgo', rangoFechaSesgo);
  f.EjecutarLectura;

  if f.version < 82 then
    setlength(guia_eg, length(guia)); // no los dejo en CERO.

  //  cantValoresDeterministicosUsados := 0;
end;


procedure TPronostico.Free;
begin
  setlength(guia, 0);
  setlength(sesgo, 0);
  setlength(factor, 0);
  if fechaIniSesgo_ <> nil then
    FechaIniSesgo_.Free;
  inherited Free;
end;


function TPronostico.determinismoSoloEstadoInicial: boolean;
begin
  Result := NPSA = 0;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TPronostico.ClassName, TPronostico);
  registrarClaseDeCosa(TPronosticos.ClassName, TPronosticos);
end;

procedure AlFinal;
begin
end;



end.

